
/*************************************************************************

  APU R 2.0

  molynomial.h

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  

#if !defined(MOLYNOMIAL_HDR) 
#define MOLYNOMIAL_HDR

#include "list.h"
#include "variable.h"
#include "field.h"

class Molynomial;
class TermNode;

/* Represents monomials as a coefficient and an array of exponents. 
   The array derives its size from the int value ``Num_vars'' defined
   as member of the Polynomial class. This ensures consistency between
   the two representations. However, if the user updates 
   Polynomial::Num_vars after a Monomial is created, it will cause
   problems. */ 
class Monomial { 
  private : 
    Field_element* feP; 
    Array<int> expons; 
  public : 
    static const Monomial zero_monomial; // the global null term list

    Monomial(); 
    Monomial(Field_element); Monomial(Field_element,Array<int>); 
    Field_element* fePtr_of() const; 
    Array<int> expons_of() const; 
    friend Molynomial poly_to_moly(const Polynomial&); 
};

inline Monomial::Monomial() { 
} 

inline Monomial::Monomial(Field_element fe) {  
  feP = new Field_element; 
  *feP = fe; 
  expons = Array<int>(Polynomial::Num_vars); 
  for (int i=0;i<Polynomial::Num_vars;i++) 
    expons[i] = 0; 
} 

inline Monomial::Monomial(Field_element fe, Array<int> int_arr) { 
  feP = new Field_element; 
  *feP = fe; 
  expons = Array<int>(Polynomial::Num_vars); 
  for (int i=0;i<size_of(int_arr);i++) 
    expons[i] = int_arr[i]; 
} 

inline Field_element* Monomial::fePtr_of() const { 
  return feP; 
}

inline Field_element* fePtr_of(const Monomial& m) { 
  return m.fePtr_of(); 
}

inline Array<int> Monomial::expons_of() const { 
  return expons; 
}

inline Array<int> expons_of(const Monomial& m) { 
  return m.expons_of(); 
}

/* Molynomial is a class defining polynomials in monomial form. The
   monomials are strung together as a linked list. */
class Molynomial { 
  private : 
    static const char* Class_name; 

    List<Monomial> monoms; // Linked list of monomials
  public : 
    static const Polynomial zero_moly; // the global zero molynomial 

    Molynomial();
    Molynomial(const Monomial&); // Create a new Moly with the monomial as
				 // only member.  
    Molynomial(const List<Monomial>&); // Create a new Moly with the monomial 
				       // list elts as members.
    void concat(const Monomial&); 
    List<Monomial> monoms_of() const; 
    void print() const; 
    void print(ostream&) const; 

    friend Molynomial concat(const Molynomial&, const Molynomial&); 
    friend int num_monoms(const Molynomial&); 
    friend Array<Field_element> coeffs_of(const Molynomial&); 
    friend Molynomial poly_to_moly(const Polynomial&); 
    friend ostream& operator << (ostream&, const Molynomial&); 
};

inline Molynomial::Molynomial() { 
}

inline Molynomial::Molynomial(const Monomial& m) { 
  monoms.add(m); 
}

inline Molynomial::Molynomial(const List<Monomial>& mlist) { 
  monoms = mlist; 
}

inline List<Monomial> Molynomial::monoms_of() const { 
  return monoms; 
}

inline List<Monomial> monoms_of(const Molynomial& moly) { 
  return moly.monoms_of(); 
}

inline void Molynomial::concat(const Monomial& m) { 
  monoms.add(m); 
}

inline void concat(Molynomial moly, Monomial m) { 
  moly.concat(m); 
}

inline void Molynomial::print() const { 
  cout << *this;
}

inline void print(const Molynomial& m) { 
  m.print(); 
}

inline Molynomial concat(const Molynomial& moly1, const Molynomial& moly2) { 
  Molynomial result = Molynomial(concat(moly1.monoms,moly2.monoms)); 
  return result; 
}

inline int num_monoms(const Molynomial& moly) { 
  ListNode<Monomial>* temp; 
  int result = 0; 
  for (temp=head_of(moly.monoms);temp!=0;temp=temp->next_of())
    result++; 
  return result; 
}

inline Array<Field_element> coeffs_of(const Molynomial& moly) { 
  int count = num_monoms(moly); 
  Array<Field_element> result = Array<Field_element>(count); 
  ListNode<Monomial>* temp; 
  int i = 0; 
  for (temp=head_of(moly.monoms);temp!=0;temp=temp->next_of()) { 
    result[i] = *((temp->item_of()).fePtr_of());
    i++;
  }
  return result; 
}

#endif
