
/*************************************************************************

  APU R 2.0

  polynomial.h

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  

#if !defined(POLYNOMIAL_HDR) 
#define POLYNOMIAL_HDR

#include "list.h"
#include "variable.h"
#include "field.h"


class Polynomial;
class TermNode;

class TermList { 
  private : 
    TermNode* tnode; 
  public : 
    static const TermList zero_termlist; // the global null term list

    TermList(); 
    TermList(TermNode); 
    TermList(TermNode*); 
    bool is_null() const; 
    int lead_degree() const; 
    Polynomial lead_coeff() const;
    TermNode* tnode_of() const; 
    TermNode* leadnodePtr_of() const;
    TermNode* last(); 
    TermList rest_of();           // return rest of termlist as TermList type
    void print(ostream&,int); 
    bool is_zerotermlist() const; 

    friend TermList operator + (const TermList&, const TermList&); 
    friend TermList operator + (const TermList&, const Field_element&);
    friend TermList operator + (const Field_element&, const TermList&);
    friend TermList operator + (const TermList&, const Polynomial&);
    friend TermList operator + (const Polynomial&, const TermList&);

    friend TermList operator - (const TermList&, const TermList&); 
    friend TermList operator - (const TermList&, const Field_element&);
    friend TermList operator - (const Field_element&, const TermList&);
    friend TermList operator - (const TermList&, const Polynomial&);
    friend TermList operator - (const Polynomial&, const TermList&);

    friend TermList operator * (const TermList&, const TermList&); 
    friend TermList operator * (const TermList&, const Field_element&);
    friend TermList operator * (const Field_element&, const TermList&);
    friend TermList operator * (const TermList&, const Polynomial&);
    friend TermList operator * (const Polynomial&, const TermList&);

    friend TermList operator / (const TermList&, const TermList&); 
    friend TermList operator / (const TermList&, const Field_element&);
    friend TermList operator / (const Field_element&, const TermList&);
    friend TermList operator / (const TermList&, const Polynomial&);
    friend TermList operator / (const Polynomial&, const TermList&);

    friend TermList operator % (const TermList&, const TermList&); 
    friend TermList operator % (const TermList&, const Field_element&);
    friend TermList operator % (const Field_element&, const TermList&);
    friend TermList operator % (const TermList&, const Polynomial&);
    friend TermList operator % (const Polynomial&, const TermList&);

    friend ostream& operator << (ostream&, const TermList&);
};


class Polynomial { 
  private : 
    static const char* Class_name; 

    Field_element* feP; // contains field elt for terminal polys, var o/w
    TermList tlist; 	// Linked list of termnodes
  public : 
    static List<Variable> var_list; 
    static int Num_vars; 
    static const Polynomial zero_poly; // the global zero polynomial 

    Polynomial();
    Polynomial(int); 
    Polynomial(unsigned int); 
    Polynomial(float); 
    Polynomial(double); 
    Polynomial(const Field_element&);// make constant polynomial
    Polynomial(const Slp&);          // make terminal poly with slp as coeff
    Polynomial(char*);               // make poly "x" where "x" is string
    Polynomial(char*,int);           // make poly "x^n" where n is degree
    Polynomial(char*, TermList);     // make poly in given var and termlist
    Polynomial(Variable);
    Polynomial(Variable*);
    Polynomial(Variable*, TermList);
    Variable* variable_of() const; 
    Field_element* fePtr_of() const;
    Field_element fe_of() const; 
    bool is_terminal () const;
    TermList termlist_of() const;
    int degree();
    int degree() const;
    Polynomial lcoeff() const; 
    void print(ostream&,int) const;
    bool is_zeropoly() const; 

    friend Polynomial operator + (const Polynomial&, const Polynomial&);
    friend Polynomial operator + (const Polynomial&, const Field_element&);
    friend Polynomial operator + (const Field_element&, const Polynomial&);
    friend Polynomial operator + (const Polynomial&, const int&);
    friend Polynomial operator + (const int&, const Polynomial&);

    friend Polynomial operator - (const Polynomial&, const Polynomial&);
    friend Polynomial operator - (const Polynomial&, const Field_element&);
    friend Polynomial operator - (const Field_element&, const Polynomial&);
    friend Polynomial operator - (const Polynomial&, const int&);
    friend Polynomial operator - (const int&, const Polynomial&);

    friend Polynomial operator - (const Polynomial&);

    friend Polynomial operator * (const Polynomial&, const Polynomial&);
    friend Polynomial operator * (const Polynomial&, const Field_element&);
    friend Polynomial operator * (const Field_element&, const Polynomial&);
    friend Polynomial operator * (const Polynomial&, const int&);
    friend Polynomial operator * (const int&, const Polynomial&);

    friend Polynomial operator / (const Polynomial&, const Polynomial&);
    friend Polynomial operator / (const Polynomial&, const Field_element&);
    friend Polynomial operator / (const Field_element&, const Polynomial&);
    friend Polynomial operator / (const Polynomial&, const int&);
    friend Polynomial operator / (const int&, const Polynomial&);

    friend Polynomial operator % (const Polynomial&, const Polynomial&);
    friend Polynomial operator % (const Polynomial&, const Field_element&);
    friend Polynomial operator % (const Field_element&, const Polynomial&);

    friend Polynomial operator ^ (const Polynomial&, int);

    friend Polynomial operator == (const Polynomial&, const Polynomial&);
    friend Polynomial operator == (const Polynomial&, const Field_element&);
    friend Polynomial operator == (const Field_element&, const Polynomial&);
    friend Polynomial operator == (const Polynomial&, const int&);
    friend Polynomial operator == (const int&, const Polynomial&);

    friend ostream& operator << (ostream&, const Polynomial&);
    friend istream& operator >> (istream&, Polynomial&);
    friend Variable* new_polyvariable(char*);
};

class TermNode { 
  private : 
    static const char* Class_name; 
    int degree; 
    Polynomial* coeff; 
    TermNode* next; 
  public : 
    TermNode();
    TermNode(int, Polynomial); 
    TermNode(int, Polynomial*); 
    TermNode* next_of(); 	// get next node
    void set_next(TermNode);	// set next node
    int degree_of();
    Polynomial* coeff_of(); 
    Polynomial coeff_poly(); 
    void set_coeff_poly(const Polynomial&); 
};

inline TermNode::TermNode() {
  degree = 0; 
  coeff = 0; 
  next = 0; 
}

inline TermNode* TermNode::next_of() { return(next); } 

inline TermList::TermList() {
  tnode = 0; 
}

inline Polynomial::Polynomial() {
  feP = 0; 
}

inline Polynomial::Polynomial(int a) {
  if (a == 0) 
    *this = zero_poly;
  else feP = new Field_element(a); 
}

inline Polynomial::Polynomial(unsigned int a) {
  if (a == (0 % IntModp::prime)) 
    *this = zero_poly;
  else feP = new Field_element(a); 
}

inline Polynomial::Polynomial(float a) {
  if (a == 0.0) 
    *this = zero_poly;
  else feP = new Field_element(a); 
}

inline Polynomial::Polynomial(double a) {
  if (a == 0.0) 
    *this = zero_poly;
  else feP = new Field_element(a); 
}

inline Polynomial::Polynomial(const Field_element& a) {
  if (is_zero(a))
    *this = zero_poly;
  else {
    feP = new Field_element; 
    *feP = a; 
  }
}

inline Polynomial::Polynomial(const Slp& slp) {
  if (is_zero(slp))
    *this = zero_poly;
  else {
    feP = new Field_element;
    *feP = (Field_element)(slp);
  }
}

inline TermList::TermList(TermNode a) {
  tnode = new TermNode; 
  *tnode = a; 
}

inline TermList::TermList(TermNode* a) {
  tnode = a; 
}

inline bool TermList::is_null() const {
  if (tnode == 0) return(1); else return(0);
}

inline int TermNode::degree_of() { 
  return(degree); 
}

inline TermNode::TermNode(int deg, Polynomial* given_coeff) { 
  if (deg < 0) cout << "Error - asked to creat TermNode with deg < 0\n";
  else {
    degree = deg; 
    coeff = given_coeff; 
    next = 0;
  }
}

inline TermNode::TermNode(int deg, Polynomial given_coeff) { 
  if (deg < 0) cout << "Error - asked to creat TermNode with deg < 0\n";
  else {
    degree = deg; 
    coeff = new Polynomial; 
    *coeff = given_coeff; 
    next = 0;
  }
}

inline int TermList::lead_degree() const {
  if (tnode == 0) return(0); else return(tnode->degree_of()); 
}

inline bool TermList::is_zerotermlist() const {
  if (tnode == 0) return(true); else return(false);
}

inline bool is_zerotermlist(const TermList& tlist) {
  return(tlist.is_zerotermlist());
}

inline Variable* new_polyvariable(char* var) { 
  Variable* temp = new Variable(var); 
  temp->index((Polynomial::Num_vars)); 
  temp->slp(Slp(temp)); 
  temp->value(Field_element((int)(lrand48())));
  Polynomial::var_list.add(*temp); 
  (Polynomial::Num_vars)++; 
  return temp; 
}

inline Polynomial::Polynomial(char* variable, TermList a) {
  Variable* temp = var_list.find(variable); 
  if (temp == 0) 
    temp = new_polyvariable(variable); 
  if (is_zerotermlist(a)) {
    *this = zero_poly;
  }
  else if (temp != 0) {
    feP = temp; 
    tlist = a; 
  }
  else cout << variable << " not in list of Polynomial variables\n"; 
}

/* Make the polynomial "x" where "x" is the variable */
inline Polynomial::Polynomial(char* variable) {
  Variable* temp = var_list.find(variable); 
  if (temp == 0) 
    temp = new_polyvariable(variable); 
  if (temp != 0) {
    feP = temp; 
    tlist = TermList(TermNode(1,Polynomial(1)));
  }
  else cout << variable << " not in list of Polynomial variables\n"; 
}

/* Make the polynomial "x^n" where "x" is the variable and n is the exponent */
inline Polynomial::Polynomial(char* variable, int n) {
  Variable* temp = var_list.find(variable); 
  if (temp == 0) 
    temp = new_polyvariable(variable); 
  if (temp != 0) {
    feP = temp; 
    tlist = TermList(TermNode(n,Polynomial(1)));
  }
  else cout << variable << " not in list of Polynomial variables\n"; 
}

inline Polynomial::Polynomial(Variable var) {
  feP = new Field_element;
  *feP = (Field_element&)(var);
  tlist = TermList(TermNode(1,Polynomial(1)));
}

inline Polynomial::Polynomial(Variable* var_ptr) {
  feP = var_ptr; 
  tlist = TermList(TermNode(1,Polynomial(1)));
}

inline Polynomial::Polynomial(Variable* var_ptr, TermList a) {
  if (is_zerotermlist(a)) {
    *this = zero_poly;
  }
  else {
    feP = var_ptr; 
    tlist = a; 
  }
}

inline Variable* Polynomial::variable_of() const {
  return((Variable*)feP); 
}

inline Field_element* Polynomial::fePtr_of() const {
  return(feP);
}

inline Field_element* fePtr_of(const Polynomial& poly) {
  return(poly.fePtr_of());
}

inline bool Polynomial::is_terminal() const {
  if (tlist.is_null()) return(true); else return(false);
}

inline TermList Polynomial::termlist_of() const {
  return(tlist); 
}

inline bool Polynomial::is_zeropoly() const {
  if (fePtr_of() == 0) return(true);
  else return(false);
}

inline Field_element Polynomial::fe_of() const {
  if (!is_zeropoly())
    return (*feP);
  else return Field_element(0);
}

inline int Polynomial::degree() {
  if (is_zeropoly()) return(0);
  else if (is_zerotermlist(termlist_of())) return(0);
  else return((termlist_of()).lead_degree()); 
}

inline int Polynomial::degree() const {
  if (is_zeropoly()) return(0);
  else if (is_zerotermlist(termlist_of())) return(0);
  else return((termlist_of()).lead_degree()); 
}

inline Polynomial* TermNode::coeff_of() { 
  return(coeff);
}

inline Polynomial TermNode::coeff_poly() { 
  return(*coeff);
}

inline void TermNode::set_coeff_poly(const Polynomial& p) { 
  if (coeff != 0) 
    *coeff = p; 
  else { coeff = new Polynomial; *coeff = p; }
}

inline Polynomial TermList::lead_coeff() const{ 
  return(*(tnode->coeff_of()));
}

inline Polynomial Polynomial::lcoeff() const {
  if (is_zeropoly()) return(0);
  else if (!is_terminal()) return(termlist_of().lead_coeff());
  else return(Polynomial(fe_of()));
}

inline TermNode* TermList::tnode_of() const {
  return(tnode); 
}

inline TermNode* tnode_of(const TermList& tl) {
  return(tl.tnode_of());
}

inline TermNode* TermList::leadnodePtr_of() const {
  return(tnode); 
}

inline TermNode* leadnodePtr_of(const TermList& tl) {
  return(tl.leadnodePtr_of());
}

inline TermNode* TermList::last() { 
  TermNode* temp = tnode; 
  while (temp->next_of() != 0) {
    temp = temp->next_of(); 
  } 
  return(temp); 
}

inline void TermNode::set_next(TermNode tn) {
  next = new TermNode; 
  *next = tn;
}

inline TermList TermList::rest_of() {
  return(TermList((tnode_of())->next_of()));
}

Polynomial prem(const Polynomial&, const Polynomial&);
List<Polynomial> sturm_habicht(const Polynomial&, const Polynomial&);
List<Polynomial> subres_prs(const Polynomial&, const Polynomial&);
Polynomial sylvester_res(const Polynomial&, const Polynomial&);

Polynomial diff (const Polynomial&, char*, int= NOTYPE); 
Polynomial diff (const Polynomial&); 
Polynomial diff(const Polynomial&, int,char*); 
Polynomial diff2 (const Polynomial&, char*); 

#include "matrix.h"
Polynomial det (const Matrix<Polynomial>&);
Matrix<Polynomial> gauss_elim (const Matrix<Polynomial>&);

#include "poly_aux.h"

/*
inline max_of() const {
  cout << "max_of() not defined for Array<Polynomial>\n";
}
*/

#endif
