
/*************************************************************************

  APU R 2.0

  poly_diff.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "polynomial.h"
#include "slp.h"
extern int Debug; 


/* Computes the derivative of the given polynomial.
   The type of derivative computed depends on the flag deriv_type.

   If deriv_type is SLP, the polynomial is assumed to have Slp's
   as its coefficients and the derivative of the coefficient
   Slp's will be taken wrt the variable var. 

   If the deriv_type is VAR, the derivative of the polynomial
   will be taken wrt to the variable var. 

   If the deriv_type is some other field type, the derivative
   of the polynomial will be taken wrt the leading variable. 
*/

Polynomial
diff (const Polynomial& poly, char* var, int deriv_type = NOTYPE) 
{
   Polynomial result;
   switch(deriv_type) {
     case VAR : result = diff2(poly,var); break;
     case SLP : result = diff(poly,SLP,var); break;
     default : result  = diff(poly); break;
   }
   return(result);
}

Polynomial 
diff (const Polynomial& poly) 
{
  if (is_zeropoly(poly))
    return(poly);
  else if (is_terminal(poly))
    return(Polynomial::zero_poly);

  TermList tl = termlist_of(poly); 
  TermNode *lead, *curr;
  lead = 0; curr = 0; 
  for (TermNode* temp = leadnodePtr_of(tl);temp!=0; temp = temp->next_of()) {
    if (lead == 0) {
      if (temp->degree_of() > 0) {
        if (!is_zeropoly(temp->coeff_poly())) {
          lead = new TermNode(temp->degree_of()-1,
			  Polynomial(temp->degree_of())*temp->coeff_poly());
          curr = lead; 
	  if (Debug) { 
	    cout << temp->coeff_poly() << "\n"; 
	    cout << temp->degree_of() << "\n"; 
	    cout << lead->coeff_poly() << "\n"; 
	  } 
        }
      }
    }
    else {
      if (temp->degree_of() > 0) {
        if (!is_zeropoly(temp->coeff_poly())) {
          curr->set_next(TermNode(temp->degree_of()-1,
			  Polynomial(temp->degree_of())*temp->coeff_poly()));
	  curr = curr->next_of();
	  if (Debug) { 
	    cout << temp->coeff_poly() << "\n"; 
	    cout << temp->degree_of() << "\n"; 
	    cout << curr->coeff_poly() << "\n"; 
	  } 
        }
      }
    }
  }
  Polynomial result = Polynomial(variable_of(poly),TermList(lead));
  return(result);
}

Polynomial
diff(const Polynomial& poly, int deriv_type, char* var)
{
  if (deriv_type != SLP) 
    return(Polynomial::zero_poly);
  else {
    if (is_zeropoly(poly))
      return(poly);
    else if (!is_terminal(poly)) {
      TermList tl = termlist_of(poly); 
      TermNode *lead, *curr;
      lead = 0; curr = 0; 
      for (TermNode* temp=leadnodePtr_of(tl);temp!=0;temp =temp->next_of()) {
	if (lead == 0) {
          lead = new TermNode(temp->degree_of(),
			      diff(temp->coeff_poly(),deriv_type,var));
	  curr = lead;
	}
	else {
          curr->set_next(TermNode(temp->degree_of(),
			      diff(temp->coeff_poly(),deriv_type,var)));
	  curr = curr->next_of();
	}
      }
      Polynomial result = Polynomial(variable_of(poly),TermList(lead));
      return(result);
    }
    else {
      if (type_of(fe_of(poly)) == SLP) {
	Slp temp_slp = diff(*((Slp*)fePtr_of(poly)),var); 
	Polynomial result = Polynomial(temp_slp);
	return(result);
      }
      else return(Polynomial::zero_poly);
    }
  }
}

Polynomial
diff2 (const Polynomial& poly, char* var) 
{
  if (is_zeropoly(poly)) 
    return(poly);
  else if (is_terminal(poly))
    return(Polynomial::zero_poly);
  else {
    if (!strcmp(variable_of(poly)->name(),var)) {
      Polynomial result = diff(poly);
      return(poly);
    }
    else {
      TermNode *lead,*curr; 
      TermList tl = termlist_of(poly);
      lead = 0; curr = 0;
      for (TermNode* temp=leadnodePtr_of(tl);temp!=0;temp=temp->next_of()) {
        if (lead == 0) {
          if (!is_zeropoly(temp->coeff_poly())) {
            lead = new TermNode(temp->degree_of(),
				diff2(temp->coeff_poly(),var));
            curr = lead; 
          }
        }
	else {
          if (!is_zeropoly(temp->coeff_poly())) {
            curr->set_next(TermNode(temp->degree_of(), 
			   diff2(temp->coeff_poly(),var)));
	    curr = curr->next_of();
          }
	}
      }
      Polynomial result = Polynomial(variable_of(poly),TermList(lead));
      return(result);
    }
  }
}
