
/*************************************************************************

  APU R 2.0

  poly_other.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "polynomial.h"
#include "list.h"

extern int Debug;

/* Computes the psuedo-remainder of dividend wrt divisor. 
   See Knuth, vol. 2 for algorithm details */

Polynomial prem_aux(const Polynomial&, const Polynomial&);
List<Polynomial> sturm_habicht(const Polynomial&, const Polynomial&); 
List<Polynomial> subres_prs(const Polynomial&, const Polynomial&); 
Polynomial sylvester_res(const Polynomial&, const Polynomial&);

Polynomial
prem(const Polynomial& dividend, const Polynomial& divisor)
{
  if (is_zeropoly(dividend)) 
    return(Polynomial::zero_poly); 
  if (is_zeropoly(divisor)) {  
    cout << "Error in prem() - prem by zero divisor\n";  
    return(Polynomial::zero_poly); 
  } 
  Polynomial result;
  if (dividend.variable_of()->index() < divisor.variable_of()->index()) {
    if (dividend.degree() > 0) 
      result = Polynomial::zero_poly; 
    else 
      result = prem(dividend.lcoeff(),divisor);
  }
  else if (dividend.variable_of()->index() > divisor.variable_of()->index()) {
    if (divisor.degree() > 0) 
      result = dividend;
    else 
      result = prem(dividend,divisor.lcoeff());
  }
  else { // Both polys in same variable - so check degrees
    if (dividend.degree() > 0) {
      if (divisor.degree() > 0) 
	result = prem_aux(dividend,divisor);
      else 
	result = Polynomial::zero_poly;
    }
    else {
      if (divisor.degree() > 0) 
	result = dividend;
      else 
	result = prem(dividend.lcoeff(),divisor.lcoeff());
    }
  }
  if (!is_zeropoly(result)) { 
    while ((degree(result) == 0) && (!is_terminal(result))) 
      result = lcoeff(result); 
  } 
  return(result); 
}

/* prem_aux assumes that the polynomials are both in the same variable
   and both have non-zero-degree leading terms in that variable */
Polynomial
prem_aux(const Polynomial& dividend, const Polynomial& divisor)
{
  Polynomial result;
  int n, m; 
  n = dividend.degree();
  m = divisor.degree();

  if (n < m)
    result = dividend;
  else {
    char* var_name = dividend.variable_of()->name(); 
    Polynomial cur_divisor;
    Polynomial cur_dividend = dividend;
    Polynomial rest_dividend = 
	   Polynomial(var_name,dividend.termlist_of().rest_of()); 
    Polynomial rest_divisor = 
	   Polynomial(var_name,divisor.termlist_of().rest_of()); 

    Polynomial temp; 
    int count = 0;

    while (n >= m) {
      if (cur_dividend.degree() == n) {
        temp = Polynomial(var_name,n-m) * cur_dividend.lcoeff();
	if (Debug) 
	  cout << temp << "\n";
        cur_divisor = temp * rest_divisor;
        cur_dividend = rest_dividend * divisor.lcoeff();
        cur_dividend = cur_dividend - cur_divisor;
	if (is_zeropoly(cur_dividend)) 
	  n = m-1;
	else 
	  rest_dividend = 
	    Polynomial(var_name,cur_dividend.termlist_of().rest_of());
      }
      else count++; 
      n--;
    }
    if (count != 0) 
      cur_dividend = cur_dividend * ((divisor.lcoeff())^count);
    result = cur_dividend;
  }
  return(result);
}

/* Returns the Sturm-Habicht seq of two polys in the form of a list of
   polynomials. Based on Maple code by Felipe Cucker - see "A Handout
   for IF". Does not, in general, return the exact Sylvester resultant
   as the last element of the sequence but rather a multiple of it.
*/

List<Polynomial>
sturm_habicht(const Polynomial& p1, const Polynomial& p2) 
{
  List<Polynomial> result;
  Polynomial s0,s1,R_coeff;
  int n,d,j,r,done;

  result.add(p1);
  result.add(p2);

  int n1 = p1.degree();
  int n2 = p2.degree();


  if (n1 > n2) {
    n = n1 - 1;
    d = n - n2;
    if (d != 0) {
      Polynomial temp = p1.lcoeff() * p2.lcoeff();
      temp = temp^d;
      if ( ((d*(d+1)/2) % 2) != 0)
        temp = - temp;
      s1 = p2 * temp;
      temp = (p1.lcoeff())^d;
      if ( ((d+2)*(d+3)/2) % 2 != 0)
	temp = - temp;
      s0 = prem(p1,p2);
      s0 = s0*temp;
      result.add(s1);
      if (!is_zeropoly(s0))
	result.add(s0);
    }
    else {
      s1 = p2;
      s0 = prem(p1,p2);
      if (Debug) 
	cout << s0 << "\n"; 
      if (!is_zeropoly(s0)) {
	s0 = -s0;
        if (Debug) 
	  cout << s0 << "\n"; 
	result.add(s0);
      }
    }
    j = n2 - 1;
  }
  else {
    n = n2;
    s1 = p2;
    Polynomial temp = (p2.lcoeff())^2; 
    temp = p1*temp;
    s0 = temp % p2;
    s0 = -s0;
    j = n-1;
    if (!is_zeropoly(s0))
      result.add(s0);
  }
  R_coeff = s1.lcoeff();
  if (is_zeropoly(s0))
    r = -1;
  else r = s0.degree();
  done = 0;
  while (r >= 0) {
    if (r > 0) {
      if (r == j) {
	Polynomial temp1 = R_coeff^(j-r+2);
	if (Debug) {
	  cout << R_coeff << "\n"; 
	  cout << temp1 << "\n"; 
	}
	if ((j-r+2) % 2 != 0) 
	  temp1 = - temp1;
	Polynomial temp2 = prem(s1,s0);
	if (Debug) 
	  cout << temp2 << "\n";
	temp2 = temp2 / temp1;

	if (((j-r+2)*(j-r+3)/2) % 2 != 0)
	  temp2 = - temp2;
	if (!is_zeropoly(temp2))
	  result.add(temp2);
	s1 = s0;
	s0 = temp2;
	R_coeff = s1.lcoeff();
	j = r-1;
      }
      else {
	Polynomial temp1 = R_coeff^(j-r+2);
	Polynomial temp2 = prem(s1,s0);
	temp2 = temp2 / temp1;
	temp1 = (s0.lcoeff())^(j-r);
	Polynomial temp3 = s0*temp1;
	s1 = temp3/(R_coeff^(j-r));
	if ( ((j-r)*(j-r+1)/2) % 2 != 0) 
	  s1 = -s1;
	result.add(s1);
	s0 = temp2;
	if ( ((j-r+2)*(j-r+3)/2) % 2 != 0) 
	  s0 = -s0;
	if (!is_zeropoly(s0))
	  result.add(s0);
	R_coeff = s1.lcoeff();
	j = r-1;
      }
    }
    else {
      if (j>0) {
	Polynomial temp1 = (s0.lcoeff())^j;
	Polynomial temp2 = s0*temp1;
	s1 = temp2/(R_coeff^j);
	if ( (j*(j+1)/2) % 2 != 0) 
	  s1 = -s1;
	result.add(s1);
      }
      done = 1;
    }
    if (!is_zeropoly(s0) && !done) 
      r = s0.degree();
    else r = -1;
  }
  return(result);
}

/* Computes the subresultant PRS according to Collins algorithm as
   given by Brown and Traub (JACM, v.18, no.4). The advantage of
   this algorithm is that it returns the exact resultant as the
   last polynomial in the sequence
*/

List<Polynomial>
subres_prs(const Polynomial& p1, const Polynomial& p2) 
{
  List<Polynomial> result;

  Polynomial p,q, A_i, A_j, A_k; 
  Polynomial psi,c_i,c_j,temp;
  int d_i,d_j, del_i, n1, n2; 

  n1 = degree(p1); n2 = degree(p2);
  if (n2 > n1) {
    q = p1; p = p2;
  }
  else {
    p = p1; q = p2; 
  }
  result.add(p);
  result.add(q);
  A_i = p;
  A_j = q;
  c_i = lcoeff(A_i);
  c_j = lcoeff(A_j);
  d_i = degree(A_i);
  d_j = degree(A_j);
  del_i = d_i - d_j; 
  A_k = prem(p,q);
  if ((del_i+1) % 2 == 1) 
    A_k = - A_k;
  psi = (-c_j)^del_i;
  if (((1-del_i) % 2 == 1) || ((1-del_i) % 2 == -1)) 
    psi = -psi; 
  A_i = A_j;
  A_j = A_k;
  c_i = lcoeff(A_i);
  c_j = lcoeff(A_j);
  d_i = degree(A_i);
  d_j = degree(A_j);
  del_i = d_i - d_j;
  while (!is_zeropoly(A_j)) {
    result.add(A_k);
    A_k = prem(A_i,A_j);
    temp = - c_i;
    temp = temp * (psi^del_i);
    A_k = A_k / temp;
    temp = (-c_j) ^ del_i;
    if (1 - del_i > 0) 
      psi = temp * (psi^(1-del_i));
    else if (1 - del_i < 0) 
      psi = temp / ( psi ^ (del_i - 1));
    else psi = temp; 
    A_i = A_j;
    A_j = A_k;
    d_i = degree(A_i); d_j = degree(A_j); 
    del_i = d_i - d_j;
    c_i = lcoeff(A_i);
    c_j = lcoeff(A_j);
  }
  return(result);
}

Polynomial 
sylvester_res(const Polynomial& p1, const Polynomial& p2)
{
  Polynomial result;
  if (is_zeropoly(p1) || is_zeropoly(p2)) 
    result = Polynomial::zero_poly;
  else {
    List<Polynomial> subres_seq = subres_prs(p1,p2);
    result = (subres_seq.back_of())->item_of();
  }
  return(result);
}
