
/*************************************************************************

  APU R 2.0

  poly_sign.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "poly_sign.h"
#include "matrix.h"
#include "primelist.h"
extern int Debug; 

/* Given a list of polynomials, computes the difference in sign alternations
   of the polynomials evaluated at -infinity and +infinity. 
   The polynomials should have coefficients which are field elements
   and not other polynomials.
   The parameter "multi_modp" if true, indicates a) that the coefficients
   of the polynomial are SLP's and b) that the SLP's should be interpolated
   using the MultiIntModp field element type. 
*/


int 
diff_in_alternations(const List<Polynomial>& polylist,int coeff_type=BIGINT,
bool multi_modp = false)
{
  if (polylist.is_empty())
    return(0);

  if (multi_modp) { 
    return(diff_in_alternations_modp(polylist)); 
  }
  int result;
  List<int> list1, list2; 

  Polynomial temp,temp_i; 
  int temp_sign; 
  ListNode<Polynomial>* ln; 
  Matrix<int> modp_coeffs; 
  Matrix<int> v; 
  Array<Slp> slp_coeff_array; 
  Array<BigInt> coeff_array; 

  ln = head_of(polylist); 
  if (coeff_type == SLP) { 
    v = Matrix<int>(1,Primes.num_primes());
    modp_coeffs = Matrix<int>(1,Primes.num_primes());
  }

  int count = 0; 
  if (coeff_type == SLP) {
    for (ln = head_of(polylist); ln != 0 ; ln = ln->next_of()) {
      count++; 
    }
    slp_coeff_array = Array<Slp>(count); 
    coeff_array = Array<BigInt>(count); 
    count = 0; 
    for (ln = head_of(polylist); ln != 0 ; ln = ln->next_of()) {
      temp = lcoeff(ln->item_of());
      if (!is_terminal(temp)) {
        cout << "Error - diff_in_alternations called with" << 
		" non-univariate polynomials\n"; 
        return(0);
      }
      if ((type_of(fe_of(temp)) == SLP)) {
        Slp slp = *((Slp*)(fePtr_of(temp)));
        slp_coeff_array[count] = slp; 
      }
      else { 
	cout << "Error in diff_in_alternations() : SLP flag set " << 
	"with non-Slp coefficients\n";
        return(0);
      }
      count++; 
    }
    coeff_array = interpolate_int_numerical(slp_coeff_array);
  }
  
  count = 0; 
  for (ln = head_of(polylist); ln != 0 ; ln = ln->next_of()) {
    if ((coeff_type == SLP) && (type_of(fe_of(temp)) == SLP)) {
      temp_sign = sign(coeff_array[count]); 
      cout << coeff_array[count] << "\n"; 
    }
    else { 
      temp = lcoeff(ln->item_of()); 
      temp_sign = sign(fe_of(temp));
    }
    list2.append(temp_sign);
    temp_sign = temp_sign*(1 - 2*(degree(ln->item_of()) % 2));
    list1.append(temp_sign); 
    count++; 
  }
  return(sign_alternations(list1) - sign_alternations(list2));

}

int 
sign_alternations (const List<int>& list) 
{
  int result = 0,temp = 0;
  ListNode<int>* ln; 
  for (ln = head_of(list); ln != 0; ln = ln->next_of()) {
    if (ln->item_of() > 0) {
      if (temp == 0) temp = 1;
      else if (temp < 0) {
	result ++;
	temp = 1;
      }
    }
    else if (ln->item_of() < 0) {
      if (temp == 0) temp = -1;
      else if (temp > 0) {
	result++;
	temp = -1;
      }
    }
  }
  return(result);
}

/* This procedure is to be used when the coefficients of the polynomials
   in the sequence are SLP's which are to be interpolated using 
   MultiIntModp's - i.e., using Chinese remaindering combined with
   the Field_element type MultiIntModp, which represents an integer
   as a set of numbers modulo different primes */

int
diff_in_alternations_modp(const List<Polynomial>& polylist)
{
  
  int result,count; 
  ListNode<Polynomial>* ln; 
  Array<Slp> slp_coeff_array; // We will put the lcoeff SLP's here.
  Array<BigInt> coeff_array;  // The result after interpolation will be here.
  Polynomial temp; 

  // First count the number of polynomials in the list (we need to
  // interpolate only the leading coeffs). 
  count = 0; 
  for (ln = head_of(polylist); ln != 0 ; ln = ln->next_of()) 
      count++; 
  slp_coeff_array = Array<Slp>(count); // Allocate the requisite no. 
  coeff_array = Array<BigInt>(count); 
  count = 0; 
  for (ln = head_of(polylist); ln != 0 ; ln = ln->next_of()) {
      temp = lcoeff(ln->item_of());
      if (!is_terminal(temp)) {
        cout << "Error - diff_in_alternations called with" << 
		" non-univariate polynomials\n"; 
        return(0);
      }
      if ((type_of(fe_of(temp)) == SLP)) {
        Slp slp = *((Slp*)(fePtr_of(temp)));
        slp_coeff_array[count] = slp; 
	if (Debug) { 
	  Slp::Eval_level++; 
	  Field_element fe = evaluate_multimodp(slp); 
	  cout << fe << "\n"; 
	}
      }
      else { 
	cout << "Error in diff_in_alternations() : SLP flag set " << 
	"with non-Slp coefficients\n";
        return(0);
      }
      count++; 
  }
  // Now we have the SLP's all in slp_coeff_array. 
  coeff_array = interpolate_multi_int_numerical(slp_coeff_array);

  // Now to count the number of sign alternations in lcoeffs.  
  List<int> list1, list2; int temp_sign; 
  count = 0; 
  for (ln = head_of(polylist); ln != 0 ; ln = ln->next_of()) {
    temp_sign = sign(coeff_array[count]); 
    list2.append(temp_sign);
    temp_sign = temp_sign*(1 - 2*(degree(ln->item_of()) % 2));
    list1.append(temp_sign); 
    count++; 
  }
  return(sign_alternations(list1) - sign_alternations(list2));
}
