
/*************************************************************************

  APU R 2.0

  poly_subs.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "poly_subs.h"

extern int Debug; 

/* Substitute variables in the given polynomial "poly" with the
   rational functions given by xnumer_array and xdenom_array.

   The ith element of each array gives the numerator and denominator
   respectively to be substituted for the ith variable,
   n is the bound on the number of variables.

   Algorithm uses recursive Horner substitution - see e.g.
   Zippel's book "Effective Polynomial Computation", p. 123 for
   details.
*/

void
subs (Polynomial poly, int n, Array<Polynomial> numer_array, 
    Array<Polynomial> denom_array, Polynomial *result_n, Polynomial* result_d)
{
  if (is_zeropoly(poly)) {
    *result_n = Polynomial::zero_poly; 
    return; 
  }
  if (is_terminal(poly)) { // poly is basically a field element
    *result_n = poly; 
    *result_d = Polynomial(1); 
    return;
  }
  // If we got here poly is not 0 or a terminal poly
  // First substitute for lcoeff of the poly

  Polynomial* coeff_n = new Polynomial; 
  Polynomial* coeff_d = new Polynomial; 
  subs(lcoeff(poly),n,numer_array,denom_array,coeff_n,coeff_d); 
  Polynomial res_n = *coeff_n; 
  Polynomial res_d = *coeff_d; 
  int d = degree(poly); 
  int ind = variable_of(poly)->index();
  if (ind >= n) 
    cout << "Error in subs() - variables are out of bounds\n"; 
  TermList tl = termlist_of(poly); 
  TermNode* tNodep = tl.tnode_of();
  if (tNodep->next_of() == 0) {
    res_n = res_n * (numer_array[ind]^d); 
    res_d = res_d * (denom_array[ind]^d); 
  }
  int curr_exp; 
  for (tNodep=tNodep->next_of();tNodep!=0;tNodep=tNodep->next_of()) {
    subs(tNodep->coeff_poly(),n,numer_array,denom_array,coeff_n,coeff_d); 
    if (Debug) {
      cout << *coeff_n << "\n"; 
      cout << *coeff_d << "\n"; 
    }
    curr_exp = tNodep->degree_of();
    res_n = res_n * (*coeff_d); 
    if (Debug) {
      cout << res_n << "\n"; 
      cout << res_d << "\n"; 
    }
    res_n = res_n * (numer_array[ind] ^ (d-curr_exp));
    res_d = res_d * (denom_array[ind] ^ (d-curr_exp));
    if (Debug) {
      cout << res_n << "\n"; 
      cout << res_d << "\n"; 
    }
    res_n = res_n + (res_d*(*coeff_n));
    res_d = res_d * (*coeff_d); 
    if (Debug) {
      cout << res_n << "\n"; 
      cout << res_d << "\n"; 
    }
  }
  *result_n = res_n; 
  *result_d = res_d; 
  if (Debug) {
    cout << res_n << "\n"; 
    cout << res_d << "\n"; 
  }
}

/* Substitute variables in the given polynomial "poly" with the
   rational functions given by xnumer_array and commond denominator
   given by denom.
*/

void
subs (Polynomial poly, int n, Array<Polynomial> numer_array, 
    Polynomial denom, Polynomial *result_n, Polynomial* result_d)
{
  int d_exp; 
  subs2(poly,n,numer_array,denom,result_n,&d_exp); 
  if (d_exp == 0) 
    *result_d = Polynomial(1);
  else {
    *result_d = denom; 
    for (int i=1;i<d_exp;i++) 
      *result_d = *result_d * denom;  
  }
}

void
subs2 (Polynomial poly, int n, Array<Polynomial> numer_array, 
    Polynomial denom, Polynomial *result_n, int* d_exp)
{
  if (is_zeropoly(poly)) {
    *result_n = Polynomial::zero_poly; 
    *d_exp = 0; 
    return; 
  }
  if (is_terminal(poly)) { // poly is basically a field element
    *result_n = poly; 
    *d_exp = 0; 
    return;
  }
  // If we got here poly is not 0 or a terminal poly
  // First substitute for lcoeff of the poly

  Polynomial* coeff_n = new Polynomial; 
  int cur_d_exp; 
  subs2(lcoeff(poly),n,numer_array,denom,coeff_n,&cur_d_exp); 
  Polynomial res_n = *coeff_n; 
  int d = degree(poly); 
  int d_exp1 = cur_d_exp; 
  int ind = variable_of(poly)->index();
  if (ind >= n) {
    cout << "Error in subs2() - variables are out of bounds\n"; 
    *result_n = Polynomial::zero_poly; 
    return;
  }
  TermList tl = termlist_of(poly); 
  TermNode* tNodep = tl.tnode_of();
  if (tNodep->next_of() == 0) {
    res_n = res_n * (numer_array[ind]^d); 
    d_exp1 = d_exp1 + d; 
  }
  int curr_exp; 
  for (tNodep=tNodep->next_of();tNodep!=0;tNodep=tNodep->next_of()) {
    curr_exp = tNodep->degree_of();
    d_exp1 = d_exp1 + d - curr_exp; 
    res_n = res_n * (numer_array[ind]^(d-curr_exp));
    subs2(tNodep->coeff_poly(),n,numer_array,denom,coeff_n,&cur_d_exp); 
    if (Debug) {
      cout << *coeff_n << "\n"; 
    }
    if (cur_d_exp > d_exp1) {
      res_n = res_n * (denom ^ (cur_d_exp - d_exp1));
      d_exp1 = cur_d_exp; 
    }
    else if (cur_d_exp < d_exp1) {
      *coeff_n = *coeff_n * (denom ^ (d_exp1 - cur_d_exp));
    }
    res_n = res_n + (*coeff_n); 
    d = curr_exp; 
    if ((tNodep->next_of() == 0) && (curr_exp > 0)) {
      curr_exp = 0;
      d_exp1 = d_exp1 + d - curr_exp; 
      res_n = res_n * (numer_array[ind]^(d-curr_exp));
    }

    if (Debug) {
      cout << res_n << "\n"; 
    }
  }
  *d_exp = d_exp1; 
  *result_n = res_n; 
  if (Debug) {
    cout << res_n << "\n"; 
  }
}
