
/*************************************************************************

  APU R 2.0

  rsolve.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "det_sign.h"

/* r_hat is the array of sign encodings corresponding to the array R 
   of number of real roots of the polynomial p for a given sign
   sequence of p_1 ... p_n */

Matrix<int>
r_hat_startup()
{
  Matrix<int> result = Matrix<int>(3,1); 

  result[0][0] = 0;
  result[1][0] = 1;
  result[2][0] = 2;
  return(result);
}

/* s_hat is the array of encodings of sturm queries corresponding to 
   the array S of actual query polynomials. */

Matrix<int>
s_hat_startup()
{
  Matrix<int> result = Matrix<int>(3,1);

  result[0][0] = 0;
  result[1][0] = 1;
  result[2][0] = 2;
  return(result);
}

/* Takes an array of sign encodings of the form [ [0,1 ...2], [1,1 ...2] ... ]
   with each element being an i-1 element array  and returns another array 
   with each element an i-element array such  that the ith element of each 
   corresponds to + (0), - (1) and 0 (2) added to each element.   */

Matrix<int>
augment_r_hat (Matrix<int> given_matrix) 
{
  int dim1 = dim1_of(given_matrix);
  int dim2 = dim2_of(given_matrix);

  Matrix<int> result = Matrix<int>(3*dim1,dim2+1); 
  for (int i=0;i<dim1;i++) {
    for (int j=0;j<dim2;j++) 
      result[i][j] = given_matrix[i][j];
    result[i][dim2] = 0;
  }
  for (int i=dim1;i<2*dim1;i++) {
    for (int j=0;j<dim2;j++) 
      result[i][j] = given_matrix[i-dim1][j];
    result[i][dim2] = 1;
  }
  for (int i=2*dim1;i<3*dim1;i++) {
    for (int j=0;j<dim2;j++) 
      result[i][j] = given_matrix[i-2*dim1][j];
    result[i][dim2] = 2;
  }
  return(result);
}

/* takes an r_hat array of encodings which correspond to column entries
   and computes the corresponding rows using the rows algorithm */
Matrix<int>
compute_rows(Matrix<int> r_hat)
{
  int dim1 = dim1_of(r_hat);
  int dim2 = dim2_of(r_hat);
  Array<int> col_list = Array<int>(dim1);

  for (int i=0;i<dim1;i++) 
    col_list[i] = make_decimal(r_hat[i]);
    /* first convert to decimal form for rows proc. */
  Array<int> temp = rows(col_list);
  Matrix<int> result = Matrix<int>(dim1,dim2);
  for (int i=0;i<dim1;i++) 
    result[i] = make_trinary(temp[i],dim2);
  return(result);
}

Matrix<int>
compute_Hmatrix(Matrix<int> rows, Matrix<int> cols)
{
  int dim,i,j;

  dim = dim1_of(cols);
  Matrix<int> result = Matrix<int>(dim,dim);

  for (i=0;i<dim;i++) 
    for (j=0;j<dim;j++) 
      result[i][j] = H_elt(make_decimal(rows[i]),make_decimal(cols[j]));

  return(result);
}

/* eliminates the entries in r_hat corresponding to the zero entries in R  */
Matrix<int>
eliminate_r_hat_zero_entries (Matrix<int> r_hat, Array<int> R)
{
  int dim = size_of(R);
  int count = 0;
  for (int i=0;i<dim;i++) {
    if (R[i] != 0) 
      count++;
  }
  Matrix<int> result = Matrix<int>(count,dim2_of(r_hat));
  count = 0;
  for (int i=0;i<dim;i++) {
    if (R[i] != 0) {
      result[count] = r_hat[i]; 
      count++;
    }
  }
  return(result);
}

/* eliminate zero entries in the array R and return reduced array */
Array<int>
eliminate_zero_entries (Array<int> R)
{
  int dim = size_of(R);
  int count = 0;
  for (int i=0;i<dim;i++) {
    if (R[i] != 0) 
      count++;
  }
  Array<int> result = Array<int>(count);
  count = 0;
  for (int i=0;i<dim;i++) {
    if (R[i] != 0) {
      result[count] = R[i]; 
      count++;
    }
  }
  return(result);
}

/* Right now we're just copying rows into the updated column and
   recomputing from scratch instead of using the previous values  */

Matrix<int>
update_sturm_queries (Matrix<int> rows)
{
  int dim1 = dim1_of(rows);
  int dim2 = dim2_of(rows);
  Matrix<int> result = Matrix<int>(dim1,dim2); 

  for (int i=0;i<dim1;i++) 
    result[i] = rows[i];
  return(result);
}
