
/*************************************************************************

  APU R 2.0

  slp_interp.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include <sys/types.h>
#include "slp.h"
#include "polynomial.h"
#include "poly_interp.h"
#include "crt.h"
#include "primelist.h"

extern float Timing(char*);

/* Interpolates the given Slp into a polynomial of the given variable
   var (i.e. univariate interpolation). The poly_type is the field
   type over which the polynomial is interpolated. 
   The optional argument var2 is to be used if the Slp contains
   DIFF nodes - in that case var2 is the variable wrt which the
   DIFF operator is applied. 
   The variable slp_form is used to determine whether the polynomial
   will be returned with coefficients which are Slp's of the coeffs
   of poly_type or just plain coefficients of poly_type. 
*/

extern int Debug; 

Polynomial
interpolate(const Slp& slp, char* var, int poly_type = INTMODP, 
	    char* var2 = "",bool slp_form = false) 
{
  Polynomial result;
  if (is_zero(slp))
    result = Polynomial::zero_poly; 
  else if (poly_type == BIGINT) { 
    result = interpolate_int(slp,var,var2,slp_form); 
  }
  else {
    Variable* temp_v = Slp::var_list.find(var); 
    if (temp_v == 0) {
      cout << "Illegal variable in Slp - " << var << "\n";
      return(Polynomial::zero_poly); 
    }
    Slp::Degree_eval_level++; 
    int degree; 
    degree = degree_in_var(slp,var); 
    int extra_i = degree+2; 
    Array<Polynomial> poly_array = Array<Polynomial>(degree+1); 
    Array<Field_element> eval_points = Array<Field_element>(degree+1);
    for (int i=1;i<=degree+1;i++) {
      Field_element fe = make_Field_element(i,poly_type); 
      Slp::Eval_level++; 
      int done = 1; 
      Field_element fe_val; int non_sing; 
      fe_val = evaluate(slp,var2,&non_sing,var,fe); 
      if (Debug) 
	cout << fe_val << "\n"; 
      if (!non_sing) 
	done = 0; 
      while (!done) {
	fe = make_Field_element(extra_i,poly_type); 
	fe_val = evaluate(slp,var2,&non_sing,var,fe); 
	if (non_sing) done = 1; 
	extra_i++;
      }
      eval_points[i-1] = fe; 
      poly_array[i-1] = Polynomial(fe_val); 
    }
    if (Debug) 
      for (int i=0;i<degree+1;i++) 
	cout << poly_array[i] << "\n"; 
    result = vanderpolate(poly_array,eval_points,degree,var,poly_type); 
  }
  return result;
}

/* Returns the degree of the given Slp in the given variable if the
   Slp were to be considered a rational function. Does not take
   into account possible cancellation */
int
degree_in_var(const Slp& slp, char* var) 
{
  int result;
  if (is_zero(slp))
    result = 0; 
  else if (slp.degree_eval_level() == Slp::Degree_eval_level) 
    result = slp.degree(); 
  else if (op(slp) == NOP_FE) {
    result = 0; 
    slp.degree(0);
    slp.degree_eval_level(Slp::Degree_eval_level);
  }
  else if (op(slp) == NOP_VAR) {
    if (!strcmp(var,name(variable_of(slp)))) {
      result = 1; 
      slp.degree(1); 
      slp.degree_eval_level(Slp::Degree_eval_level);
    }
    else {
      result = 0; 
      slp.degree(0); 
      slp.degree_eval_level(Slp::Degree_eval_level);
    }
  }
  else if (op(slp) == DIFF) {
    result = 0; 
    slp.degree(0); 
    slp.degree_eval_level(Slp::Degree_eval_level);
  }
  else {
    int ldeg, rdeg;
    Field_element rval; 
    ldeg = degree_in_var(lchild(slp),var); 
    if (op(slp) != EXP) 
      rdeg = degree_in_var(rchild(slp),var); 
    switch(op(slp)) {
      case ADD : result = max(ldeg,rdeg); break;
      case SUB : result = max(ldeg,rdeg); break;
      case MULT : result = ldeg + rdeg; break;
      case DIV : result = ldeg - rdeg; break;
      case EXP : 
	rval = value(rchild(slp)); 
	result = ldeg * ((SmallInt&)(rval)).i_value();
	break;
      default : cout << "Error in degree_in_var(Slp) - unknown operator\n";
    }
    slp.degree(result);
    slp.degree_eval_level(Slp::Degree_eval_level);
  }
  return result;
}

int
count_nodes(const Slp& slp) 
{
  int result;
  if (is_zero(slp))
    result = 0; 
  else if (slp.degree_eval_level() == Slp::Degree_eval_level) 
    result = 0; 
  else if (op(slp) == NOP_FE) {
    result = 1; 
    slp.degree(1);
    slp.degree_eval_level(Slp::Degree_eval_level);
  }
  else if (op(slp) == NOP_VAR) {
      result = 1; 
      slp.degree(1); 
      slp.degree_eval_level(Slp::Degree_eval_level);
  }
  else if (op(slp) == DIFF) {
    result = 2; 
    slp.degree(2); 
    slp.degree_eval_level(Slp::Degree_eval_level);
  }
  else {
    int ldeg, rdeg;
    Field_element rval; 
    ldeg = count_nodes(lchild(slp)); 
    if (op(slp) != EXP) 
      rdeg = count_nodes(rchild(slp)); 
    switch(op(slp)) {
      case ADD : result = ldeg+rdeg+1; break;
      case SUB : result = ldeg+rdeg+1; break;
      case MULT : result = ldeg + rdeg + 1; break;
      case DIV : result = ldeg + rdeg + 1; break;
      case EXP : result = ldeg + 2; break;
      default : cout << "Error in count_nodes(Slp) - unknown operator\n";
    }
    slp.degree(result);
    slp.degree_eval_level(Slp::Degree_eval_level);
  }
  return result;
}

/* This routine is used for interpolating polynomials with BigInt
   coefficients. Vandermonde interpolation will not work since
   that requires division and will therefore require rational
   numbers. Instead, we use the Chinese remainder theorem.  
*/

Polynomial
interpolate_int(const Slp& slp, char* var, char* var2 = "",
		bool slp_form = false) 
{
  Polynomial result;
  if (is_zero(slp))
    result = Polynomial::zero_poly; 
  else {
    Variable* temp_v = Slp::var_list.find(var); 
    if (temp_v == 0) {
      cout << "Illegal variable in Slp - " << var << "\n";
      return(Polynomial::zero_poly); 
    }
    Slp::Degree_eval_level++; 
    int degree = degree_in_var(slp,var); 
    cout << "Degree of interpolating polynomial = " << degree << "\n"; 

    // This is the set up for efficient memory managing for the 
    // vanderpolate_mem() routine
    Array<Polynomial> poly_array = Array<Polynomial>(degree+1); 
    Array<Field_element> eval_points = Array<Field_element>(degree+1);
    Matrix<int> poly_modp_coeffs = Matrix<int>(degree+1,Primes.num_primes());  
    Array<Polynomial> master_coeffs = Array<Polynomial>(degree+2); 
    Array<Field_element> mp_coeffs = Array<Field_element>(degree+2); 
    Polynomial master_poly; 
    Polynomial Qi_sym = get_sym_quotient_poly(var,degree,INTMODP); 
    Array<Field_element> coeff_array = Array<Field_element>(degree+1); 

    // This is the set up for efficient memory managing for the 
    // Chinese remainders routine. The array poly_coeffs stores the 
    // partial sums computed in Chinese remaindering. The array 
    // v has the "v" results in the algorithm in Knuth, v.2, p.274;
    // done_crt keeps track of which numbers are fully interpolated so far 
    // and the variable howmany_done tells how many these are, 
    // curr_primeprod is the product of the primes used till the current  
    // point; prime_index is the index of the prime used currently.  
    Array<BigInt> poly_coeffs = Array<BigInt>(degree+1); 
    Matrix<int> v = Matrix<int>(degree+1,Primes.num_primes()); 
    Array<bool> done_crt = Array<bool>(degree+1); 
    Array<BigInt> partial_sums = Array<BigInt>(degree+1); 
    int howmany_done = 0; 
    BigInt partial_sum,range,curr_primeprod,crt_result; 
    int prime_index = 0; 

    for (int j=0;j<degree+1;j++) 
      done_crt[j] = false; 

    int Done = 0; 
    int all_correct = 1; 
    int old_prime = IntModp::prime; 
    while (!Done) { 
      IntModp::prime = Primes.primes[prime_index];
      int extra_i = degree+2; 
      for (int i=1;i<=degree+1;i++) {
        Field_element fe = make_Field_element(i,INTMODP); 
        Slp::Eval_level++; 
        int done = 1; 
        Field_element fe_val; int non_sing; 
        fe_val = evaluate_modp(slp,var2,&non_sing,var,fe); 
        if (Debug) 
	  cout << fe_val << "\n"; 
        if (!non_sing) 
	  done = 0; 
        while (!done) { 
	  // If the Slp is singular when evaluated at the given point
	  // we don't bother with differentiation using L'Hopital's rule -
	  // instead we just try another number. 
	  fe = make_Field_element(extra_i,INTMODP); 
	  Slp::Eval_level++; 
	  fe_val = evaluate_modp(slp,var2,&non_sing,var,fe); 
	  if (non_sing) done = 1; 
	  extra_i++;
        }
        if (Debug) 
	  cout << fe_val << "\n"; 
        eval_points[i-1] = fe; 
        poly_array[i-1] = Polynomial(fe_val); 
      }
      if (Debug) 
        for (int i=0;i<degree+1;i++) 
	  cout << poly_array[i] << "\n"; 
      // result = vanderpolate(poly_array,eval_points,degree,var,INTMODP); 

      if (Debug) { 
	for (int i=0;i<degree+1;i++) 
	  cout << eval_points[i] << "\n"; 
      }
      // Allocates memory for the interpolation polynomials etc. These
      // are then re-used intead of being allocated anew. 
      get_mp_coeffs_mem(mp_coeffs,degree,INTMODP,eval_points); 
      if (Debug) { 
	for (int i=0;i<degree+1;i++) 
	  cout << mp_coeffs[i] << "\n"; 
      }
      get_masterpoly_coeffs_mem(master_coeffs,degree,mp_coeffs); 
      get_master_poly_mem(&master_poly,master_coeffs,degree,var); 


      result = vanderpolate_mem2(poly_array,eval_points,degree,var,
	&master_poly,&Qi_sym,coeff_array,INTMODP); 
      if (Debug) 
	cout << result << "\n"; 
      if (Debug) {
        result = vanderpolate_mem(poly_array,eval_points,degree,var,INTMODP); 
	cout << result << "\n"; 
      }
      TermList tl = termlist_of(result); 
      TermNode* tn;
      curr_primeprod = Primes.prime_prod(prime_index); 
      range = curr_primeprod/BigInt(2); 
      tn = tl.tnode_of(); 

      for (int j=degree;j>=0;j--) { 
        if ((tn == 0) || (tn->degree_of() != j)) { 
	  poly_modp_coeffs[j][prime_index] = 0; 
        }
        else {
	  if (!is_terminal(tn->coeff_poly())) { 
	    cout << "Error in interpolate_int() - non terminal poly\n"; 
	    return Polynomial::zero_poly; 
	  }
	  else {
	     poly_modp_coeffs[tn->degree_of()][prime_index] 
	       = (fe_of(tn->coeff_poly())).i_value(); 
	     tn = tn->next_of(); 
	  }
	}
	if (!done_crt[j]) { 
	  partial_sum = crt(poly_modp_coeffs[j], 
	    v[j], prime_index, partial_sums[j]); 
	  if (range <= partial_sum) 
	    crt_result = partial_sum - curr_primeprod; 
          else crt_result = partial_sum; 
	  if (Debug) { 
	    cout << poly_modp_coeffs[j][prime_index] << "\n"; 
	    cout << poly_coeffs[j] << "\n"; 
	    cout << crt_result << "\n"; 
	  }
	  if (crt_result == poly_coeffs[j]) {
	    done_crt[j] = true; 
	    howmany_done++; 
	  }
	  poly_coeffs[j] = crt_result; 
	  partial_sums[j] = partial_sum; 
        } 
      }
      if (howmany_done == degree+1)
	Done = 1; 
      prime_index++; 
      if (prime_index >= Primes.num_primes()) { 
	cout << "Primes exhausted - add more primes in file crt.c\n";
	return(Polynomial::zero_poly); 
      }
    }
    int i; 
    for (i=degree;i>=0;i--) {
      if (poly_coeffs[i] != 0) 
	break; 
    }
    if (i < 0) result = Polynomial::zero_poly; 
    else { 
      TermList tl;
      if (slp_form)
        tl = TermList(TermNode(i,Polynomial(Slp(poly_coeffs[i])))); 
      else
        tl = TermList(TermNode(i,Polynomial(poly_coeffs[i]))); 
      result = Polynomial(var,tl); 
      TermNode* tn = tl.last(); 
      for (int j=i-1;j>=0;j--) { 
	if (poly_coeffs[j] != 0) {
	  if (slp_form)
	    tn->set_next(TermNode(j,Polynomial(Slp(poly_coeffs[j]))));
	  else 
	    tn->set_next(TermNode(j,Polynomial(poly_coeffs[j])));
	  tn = tn->next_of(); 
	}
      }
    }
    IntModp::prime = old_prime;
    cout << "Primes used = " << prime_index << "\n"; 
  }
  return result;
}

/* This routine is used for interpolating Slp's which are purely numerical. 
   That is, the Slp's represent a very large number which would be
   too expensive to compute using BigInt's. This routine is used by
   diff_in_alternations() which needs the signs of numbers reprsented
   by Slp's. The routine is memory efficient because the allocation
   of the array representing the modp values of the number is
   done outside the routine. The same array is used repeatedly for
   other computations. 
*/


Polynomial
interpolate_int_numerical(const Slp& slp, char* var, 
  Matrix<int> poly_modp_coeffs,Matrix<int> v)
{
  Polynomial result;
  if (is_zero(slp))
    result = Polynomial::zero_poly; 
  else {
    Variable* temp_v = Slp::var_list.find(var); 
    if (temp_v == 0) {
      cout << "Illegal variable in Slp - " << var << "\n";
      return(Polynomial::zero_poly); 
    }
    Slp::Degree_eval_level++; 
    int degree = degree_in_var(slp,var); 

    // This is the set up for efficient memory managing for the 
    // vanderpolate_mem() routine
    Array<Polynomial> poly_array = Array<Polynomial>(degree+1); 
    Array<Field_element> eval_points = Array<Field_element>(degree+1);
    Array<Polynomial> master_coeffs = Array<Polynomial>(degree+2); 
    Array<Field_element> mp_coeffs = Array<Field_element>(degree+2); 
    Polynomial master_poly; 
    Polynomial Qi_sym = get_sym_quotient_poly(var,degree,INTMODP); 
    Array<Field_element> coeff_array = Array<Field_element>(degree+1); 

    // This is the set up for efficient memory managing for the 
    // Chinese remainders routine. The array poly_coeffs stores the 
    // partial sums computed in Chinese remaindering. The array 
    // v has the "v" results in the algorithm in Knuth, v.2, p.274;
    // done_crt keeps track of which numbers are fully interpolated so far 
    // and the variable howmany_done tells how many these are, 
    // curr_primeprod is the product of the primes used till the current  
    // point; prime_index is the index of the prime used currently.  
    if (Debug) 
      cout << sbrk(0) << "\n"; 
    Array<BigInt> poly_coeffs = Array<BigInt>(degree+1); 
    Array<bool> done_crt = Array<bool>(degree+1); 
    Array<BigInt> partial_sums = Array<BigInt>(degree+1); 
    if (Debug) 
      cout << sbrk(0) << "\n"; 
    int howmany_done = 0; 
    BigInt partial_sum,range,curr_primeprod,crt_result; 
    int prime_index = 0; 

    for (int j=0;j<degree+1;j++) 
      done_crt[j] = false; 

    int Done = 0; 
    int all_correct = 1; 
    int old_prime = IntModp::prime; 
    while (!Done) { 
      IntModp::prime = Primes.primes[prime_index];
      int extra_i = degree+2; 
      for (int i=1;i<=degree+1;i++) {
        Field_element fe = make_Field_element(i,INTMODP); 
        Slp::Eval_level++; 
        int done = 1; 
        Field_element fe_val; int non_sing = 1; 
        if (Debug) 
          cout << sbrk(0) << "\n"; 
        fe_val = evaluate_modp(slp,"",&non_sing,var,fe); 
        if (Debug) 
          cout << sbrk(0) << "\n"; 
        if (Debug) 
	  cout << fe_val << "\n"; 
        if (!non_sing) 
	  done = 0; 
        while (!done) { 
	  // If the Slp is singular when evaluated at the given point
	  // we don't bother with differentiation using L'Hopital's rule -
	  // instead we just try another number. 
	  fe = make_Field_element(extra_i,INTMODP); 
	  fe_val = evaluate(slp,"",&non_sing,var,fe); 
	  if (non_sing) done = 1; 
	  extra_i++;
        }
        eval_points[i-1] = fe; 
        poly_array[i-1] = Polynomial(fe_val); 
      }
      if (Debug) 
        for (int i=0;i<degree+1;i++) 
	  cout << poly_array[i] << "\n"; 
      // result = vanderpolate(poly_array,eval_points,degree,var,INTMODP); 

      // Allocates memory for the interpolation polynomials etc. These
      // are then re-used intead of being allocated anew. 
        if (Debug) 
          cout << sbrk(0) << "\n"; 
      get_mp_coeffs_mem(mp_coeffs,degree,INTMODP,eval_points); 
      if (Debug) { 
	for (int i=0;i<degree+1;i++) 
	  cout << mp_coeffs << "\n"; 
      }
      get_masterpoly_coeffs_mem(master_coeffs,degree,mp_coeffs); 
      get_master_poly_mem(&master_poly,master_coeffs,degree,var); 
        if (Debug) 
          cout << sbrk(0) << "\n"; 

      result = vanderpolate_mem2(poly_array,eval_points,degree,var,
	&master_poly,&Qi_sym,coeff_array,INTMODP); 
        if (Debug) 
          cout << sbrk(0) << "\n"; 
      if (Debug) 
	cout << result << "\n"; 
      if (Debug) {
        result = vanderpolate_mem(poly_array,eval_points,degree,var,INTMODP); 
	cout << result << "\n"; 
      }
      TermList tl = termlist_of(result); 
      TermNode* tn;
      curr_primeprod = Primes.prime_prod(prime_index); 
      range = curr_primeprod/BigInt(2); 
      tn = tl.tnode_of(); 

      for (int j=degree;j>=0;j--) { 
        if ((tn == 0) || (tn->degree_of() != j)) { 
	  poly_modp_coeffs[j][prime_index] = 0; 
        }
        else {
	  if (!is_terminal(tn->coeff_poly())) { 
	    cout << "Error in interpolate_int() - non terminal poly\n"; 
	    return Polynomial::zero_poly; 
	  }
	  else {
	     poly_modp_coeffs[tn->degree_of()][prime_index] 
	       = (fe_of(tn->coeff_poly())).i_value(); 
	     tn = tn->next_of(); 
	  }
	}
	if (!done_crt[j]) { 
          if (Debug) 
            cout << sbrk(0) << "\n"; 
	  partial_sum = crt(poly_modp_coeffs[j], 
	    v[j], prime_index, partial_sums[j]); 
          if (Debug) 
            cout << sbrk(0) << "\n"; 
	  if (range <= partial_sum) 
	    crt_result = partial_sum - curr_primeprod; 
          else crt_result = partial_sum; 
	  if (Debug) { 
	    cout << poly_modp_coeffs[j][prime_index] << "\n"; 
	    cout << poly_coeffs[j] << "\n"; 
	    cout << crt_result << "\n"; 
	  }
	  if (crt_result == poly_coeffs[j]) {
	    done_crt[j] = true; 
	    howmany_done++; 
	  }
	  poly_coeffs[j] = crt_result; 
	  partial_sums[j] = partial_sum; 
          if (Debug) 
            cout << sbrk(0) << "\n"; 
        } 
      }
      if (howmany_done == degree+1)
	Done = 1; 
      prime_index++; 
      if (prime_index >= Primes.num_primes()) { 
	cout << "Primes exhausted - add more primes in file crt.c\n";
	return(Polynomial::zero_poly); 
      }
    }
    int i; 
    for (i=degree;i>=0;i--) {
      if (poly_coeffs[i] != 0) 
	break; 
    }
    if (i < 0) result = Polynomial::zero_poly; 
    else { 
      TermList tl;
      tl = TermList(TermNode(i,Polynomial(poly_coeffs[i]))); 
      result = Polynomial(var,tl); 
      TermNode* tn = tl.last(); 
      for (int j=i-1;j>=0;j--) { 
	if (poly_coeffs[j] != 0) {
	  tn->set_next(TermNode(j,Polynomial(poly_coeffs[j])));
	  tn = tn->next_of(); 
	}
      }
    }
    IntModp::prime = old_prime;
  }
  return result;
}

Array<BigInt>
interpolate_int_numerical(Array<Slp> slp_array)
{
    int prev_sbrk; 
    int num_slps = size_of(slp_array); 


    // ressult contains the BigInt results of applying the Chinese rem.
    // thm. to each Slp, done_crt keeps track of which numbers are done
    // i.e. passed the CRT test, partial_sums keeps the partial sums 
    // computed so far for each Slp; modp_vals is the values mod different
    // primes for each Slp; the v matrix is the same "v" as in Knuth, vol 2
    // p. 274. 
    Array<BigInt> result = Array<BigInt>(num_slps); 
    Array<bool> done_crt = Array<bool>(num_slps); 
    Array<BigInt> partial_sums = Array<BigInt>(num_slps); 
    Matrix<int> modp_vals = Matrix<int>(num_slps,Primes.num_primes()); 
    Matrix<int> v = Matrix<int>(num_slps,Primes.num_primes()); 
    int howmany_done = 0; 
    BigInt range,curr_primeprod; 
    int prime_index = 0; 
    Field_element fe_val; 

    for (int j=0;j<num_slps;j++) 
      done_crt[j] = false; 

    prev_sbrk = sbrk(0); 
    int Done = 0; 
    int all_correct = 1; 
    int old_prime = IntModp::prime; 
    BigInt temp_2 = BigInt(2); 
    while (!Done) { 
      IntModp::prime = Primes.primes[prime_index];
      Slp::Eval_level++; 
      curr_primeprod = Primes.prime_prod(prime_index); 
      range = curr_primeprod/temp_2; 
	// cout << (sbrk(0) - prev_sbrk)/1024 << "\n"; 
	// prev_sbrk = sbrk(0); 
      for (int i=0;i<num_slps;i++) {
	BigInt partial_sum,crt_result; 
	if (!done_crt[i]) {
          fe_val = evaluate_modp(slp_array[i]); 
          if (Debug) { 
	    cout << (sbrk(0) - prev_sbrk)/1024 << "\n"; 
	    prev_sbrk = sbrk(0); 
          }
	  if (Debug) 
	    cout << fe_val << "\n"; 
          modp_vals[i][prime_index] = fe_val.i_value(); 
	  crt_mem(&partial_sum,modp_vals[i],v[i], 
			    prime_index, partial_sums[i]); 
          if (Debug) { 
	    cout << (sbrk(0) - prev_sbrk)/1024 << "\n"; 
	    prev_sbrk = sbrk(0); 
          }
	  if (range <= partial_sum) 
	    crt_result = partial_sum - curr_primeprod; 
          else crt_result = partial_sum; 
	  if (Debug) { 
	    cout << result[i] << "\n"; 
	    cout << crt_result << "\n"; 
	    cout << partial_sum << "\n"; 
	    cout << v[i][prime_index] << "\n"; 
	  }
	  if (crt_result == result[i]) {
	    done_crt[i] = true; 
	    howmany_done++; 
	  }
	  else { 
	    result[i] = crt_result; 
	    partial_sums[i] = partial_sum; 
	  }
	}
          if (Debug) { 
	    cout << (sbrk(0) - prev_sbrk)/1024 << "\n"; 
	    prev_sbrk = sbrk(0); 
          }
      }
      if (howmany_done == num_slps)
	Done = 1; 
      prime_index++; 
      if (prime_index >= Primes.num_primes()) { 
	cout << "Primes exhausted - add more primes in file crt.c\n";
	return(result); 
      }
    }
    cout << "Number of primes used = " << prime_index << "\n";  
    IntModp::prime = old_prime;
    return result;
}
