
/*************************************************************************

  APU R 2.0

  slp_interp.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include <sys/types.h>
#include "slp.h"
#include "polynomial.h"
#include "poly_interp.h"
#include "crt.h"
#include "primelist.h"

extern int Debug; 

/* This routine is used for interpolating Slp's which are purely numerical. 
   That is, the Slp's represent a very large number which would be
   too expensive to compute using BigInt's. This routine is used by
   diff_in_alternations_modp() which needs the signs of numbers reprsented
   by Slp's. The Slp's are interpolated using the MultiIntModp field
   element type, which allows us to use one Slp evaluation call to 
   compute the value mod various primes. Thus, we can reduce the total
   number of evaluation calls and reduce (recursive) function call overhead. 
*/

Array<BigInt>
interpolate_multi_int_numerical(Array<Slp> slp_array)
{
    int num_slps = size_of(slp_array); 


    // result contains the BigInt results of applying the Chinese rem.
    // thm. to each Slp, done_crt keeps track of which numbers are done
    // i.e. passed the CRT test, partial_sums keeps the partial sums 
    // computed so far for each Slp; modp_vals is the values mod different
    // primes for each Slp; the v matrix is the same "v" as in Knuth, vol 2
    // p. 274. 
    Array<BigInt> result = Array<BigInt>(num_slps); 
    Array<bool> done_crt = Array<bool>(num_slps); 
    Array<BigInt> partial_sums = Array<BigInt>(num_slps); 
    Matrix<int> modp_vals = Matrix<int>(num_slps,Primes.num_primes()); 
    Matrix<int> v = Matrix<int>(num_slps,Primes.num_primes()); 
    int howmany_done = 0; 
    BigInt range,curr_primeprod; 
    int prime_index = 0; 
    Field_element fe_val; 

    for (int j=0;j<num_slps;j++) 
      done_crt[j] = false; 

    int Done = 0; 
    int all_correct = 1; 
    int old_prime = IntModp::prime; 
    BigInt temp_2 = BigInt(2); 
    BigInt partial_sum,crt_result; 
    Slp::Eval_level++; 
    while (!Done) { 
      if (Debug) { 
	cout << prime_index << "\n"; 
      }
      curr_primeprod = Primes.prime_prod(prime_index); 
      range = curr_primeprod/temp_2; 
      for (int i=0;i<num_slps;i++) { 
        if (prime_index % MULTI_NUMPRIMES == 0) { 
	  if (!done_crt[i]) {  
            fe_val = evaluate_multimodp(slp_array[i]); 
	    for (int j=0;j<MULTI_NUMPRIMES;j++) 
	      modp_vals[i][prime_index+j] = fe_val.i_value(j); 
	  }
        }
	if (!done_crt[i]) { 
	    crt_mem(&partial_sum,modp_vals[i],v[i], 
			    prime_index, partial_sums[i]); 
	    if (range <= partial_sum) 
	        crt_result = partial_sum - curr_primeprod; 
            else crt_result = partial_sum; 
	    if (Debug) { 
	      cout << i << "\n"; 
	      cout << modp_vals[i][prime_index] << "\n"; 
	      cout << result[i] << "\n"; 
	      cout << crt_result << "\n"; 
	    }
	    if (crt_result == result[i]) {
	        done_crt[i] = true; 
	        howmany_done++; 
	    }
	    else { 
	        result[i] = crt_result; 
	        partial_sums[i] = partial_sum; 
	    }
	}
      }
      if (howmany_done == num_slps)
  	  Done = 1; 
      prime_index++; 
      if (prime_index >= Primes.num_primes()) { 
	  cout << "Primes exhausted - add more primes in file primelist\n";
	  return(result); 
      }
      if (prime_index % MULTI_NUMPRIMES == 0) { 
        Slp::Eval_level++; 
	for (int i=0;i<MULTI_NUMPRIMES;i++) 
	  MultiIntModp::primes[i] = Primes.primes[prime_index+i]; 
      }
    }
    cout << "Number of primes used = " << prime_index << "\n";  
    return result;
}
