
/*************************************************************************

  APU R 2.0

  chull.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "chull.h"
extern int Debug; 

/* Returns the points of the convex hull of the given points. Returned
   data structure is an ``Array'' (see array.h) which also has its
   size as a member so you can get the number of points from there. 
   
   Algorithm used is to test whether each point is (or is not) on
   the hull via linear programming */ 

Array<IntPoint> 
chull_points(Array<IntPoint> given_pts) 
{ 
  int num_pts = size_of(given_pts); 

  Array<int> good_indices = Array<int>(num_pts); 

  /* simp_mat is a matrix of doubles used by the simplex routine */
  int mat_size = num_pts+size_of(given_pts[0])+5; 
  double** simp_mat = new (double*)[mat_size]; 
  for (int i=0;i<mat_size;i++)
    simp_mat[i] = new (double)[num_pts+5]; 

  int count = 0; 
  for (int i=0;i<num_pts;i++) { 
    if (on_hull(given_pts[i],given_pts,simp_mat)) { 
      good_indices[count] = 1; 
      count++; 
    }
  }
  Array<IntPoint> result = Array<IntPoint>(count); 
  int j = 0; 
  for (int i=0;i<num_pts;i++) { 
    if (good_indices[i] == 1) { 
      result[j] = given_pts[i]; 
      j++; 
      if (j > count) 
	cout << "Error in chull_points()\n"; 
    }
  }
  return result; 
}

/* Memory efficient version of chull_points(), the simplex matrix is
   pre-allocated outside the routine and can be re-used */ 
Array<IntPoint> 
chull_points_mem(Array<IntPoint> given_pts,double** simp_mat) 
{ 
  int num_pts = size_of(given_pts); 

  Array<int> good_indices = Array<int>(num_pts); 

  int count = 0; 
  for (int i=0;i<num_pts;i++) { 
    if (on_hull(given_pts[i],given_pts,simp_mat)) { 
      good_indices[count] = 1; 
      count++; 
    }
  }
  Array<IntPoint> result = Array<IntPoint>(count); 
  int j = 0; 
  for (int i=0;i<num_pts;i++) { 
    if (good_indices[i] == 1) { 
      result[j] = given_pts[i]; 
      j++; 
      if (j > count) 
	cout << "Error in chull_points()\n"; 
    }
  }
  return result; 
}

bool
on_hull(IntPoint test_pt, Array<IntPoint> all_pts, double** mat)
{
  return(is_edge(test_pt,test_pt,all_pts,mat)); 
}

/* Checks whether the edge defined by head and tail is a hull edge on 
   the convex hull of all_pts. In the case head = tail will determine
   if the point is on the hull. Returns true or false. 

   The double float matrix ``mat'' is to be allocated outside is_edge() to 
   be of size (dim + num_pts + 5) x (num_pts+5). 

   (Read on if you care to know how this is done.)

   Algorithm - Convert to a linear programming problem as follows:
   Let points be: { (a_11,a_12,...), {a_21,a_22,...) ... }  
   Let head = (h_11,h_12,...) and tail = (t_11,t_12,...) 
   Lift the problem to dimension one higher i.e. d+1. The (d+1)'st 
   coordinate of all points is 1. 
   Objective function: x_1 + ... x_{d+1}. 
   Constraints: 
     0 <= x_i <= 2; for 1 <= i <= d;  
     0 <= x_{d+1} <= 2*MAXD; 
     MAXD + LEPS + a_11 + a_12 ... <= a_11 * x_1 + a_12 * x_2 + ...
     MAXD + LEPS + a_21 + a_22 ... <= a_21 * x_1 + a_22 * x_2 + ...
     ...
     MAXD + h_11 + h_12 ... = h_11 * x_1 + h_12 * x_2 + ...
     MAXD + t_11 + t_12 ... = t_11 * x_1 + t_12 * x_2 + ...
     
   Explanation of constraints: Start with the points in d-space. 
   Lift them by setting 1 as the (d+1)'st coordinate. The
   plane perpendicular to a point (a_11,a_12,...,1) is given by: 
   a_11 * x_1 + a_12 * x_2 + ... 1 * x_{d+1}. Shift this plane
   by the amount MAXD + a_11 + a_12 ... + a_1d. The common
   solution to all such planes is given by (1,1,...,1,MAXD) (assuming
   there are sufficient such planes which there will be if the object
   is of codimension 1 in the lifted space). This step ensures
   (a) the constraining planes all lie in the positive quadrant
   as required by LP (b) the planes corresponding to the points
   will get effectively ordered by whether the points are on the
   hull. (Try this with points in 1 dimension lifted to two to see
   how it works.)

*/

bool
is_edge(IntPoint head, IntPoint tail, Array<IntPoint> all_pts, double** mat)
{
  int dim = size_of(head) + 1; 
  int j; 
  for (int i=0;i<=dim;i++) { 
    mat[1][i+1] = 1.0; 
    mat[i+1][1] = 2.0; 
    for (j=1;j<=dim;j++) 
      mat[i+1][j+1] = (double)(-(i==j)); 
  }
  mat[1][1] = 0.0; 
  mat[dim+1][1] = 2.0 * MAXD; 

  int num_pts = size_of(all_pts); 

  int row = 1; 

  for (int i=0;i<num_pts;i++) {
    if (!same_point(head,all_pts[i]) && !same_point(tail,all_pts[i])) { 
      row++; 
      update_simpmat(mat,dim+row,all_pts[i],MAXD+LEPS); 
    }
  }
  update_simpmat(mat,dim+row+1,head,MAXD); 
  update_simpmat(mat,dim+num_pts+1,head,MAXD); 
  if (Debug) { 
    for (int i=0;i<dim+num_pts+4;i++) { 
      for (int j2=0;j2<num_pts+5;j2++) 
        cout << mat[i][j2] << " "; 
      cout << "\n"; 
    }
  } 
  int icase,izrov[LP_SIZE],iposv[LP_SIZE]; 
  simplx(mat,dim+num_pts,dim,dim,row-1,num_pts-row+1,&icase,izrov,iposv); 

  if (icase == -1) 
    return false; 
  else if (icase == 0) 
    return true; 
  else if (icase == 1) {
    cout << "In is_edge() - icase = 1, unbounded obj. function\n"; 
    return false; 
  }
}

void
update_simpmat(double** mat, int row, IntPoint pt, double offset) 
{ 
  int dim = size_of(pt)+1; 
  int tot = 0; 
  for (int j=0;j<dim-1;j++) { 
    mat[row][j+2] = (double)(-pt[j]); 
    tot += pt[j]; 
  }
  mat[row][dim+1] = -1.0; 
  mat[row][1] = tot + offset; 
}


/* Determines the convex hull edges of the given array of points and puts
   them in the array of edges ``result''; edge_count is a ptr to the number
   of edges found. Array<Edge> result must be preallocated by the
   calling routine - use n^2 as the upper bound, n = num of pts. 
   The matrix ``mat'' of doubles must also be preallocated of
   size (dim + num_pts + 5) x (num_pts+5). 
*/

void
hull_edges(Array<IntPoint> given_pts, Array<Edge> result, int* edge_count,
	   double** mat) 
{
  int num_pts = size_of(given_pts); 
  int count = 0; 
  for (int j=0;j<num_pts;j++) { 
    for (int k=j+1;k<num_pts;k++) { 
      if (is_edge(given_pts[j],given_pts[k],given_pts,mat)) { 
	result[count] = Edge(given_pts[j],given_pts[k]); 
	result[count].set_index(count); 
	count++; 
      }
    }
  } 
  *edge_count = count;
}
