
/*************************************************************************

  APU R 2.0

  mixvol.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "mixvol.h"
#include "mixvol_edges.h"
extern int Debug; 

/* Returns the mixed volume of the given list of polynomials. If the
   optional flag "insert" is set to be true the structures of the mixed 
   cells as arrays of edges is computed and stored - see "mixvol_edges.h" 
   for details. */
int
mixvol(List<Polynomial> poly_list,bool insert=false) { 
  List<Molynomial> moly_list; 
  ListNode<Polynomial>* temp; 
  // First convert all polys to monomial form (see "molynomial.h"). 
  for (temp=head_of(poly_list);temp!=0;temp=temp->next_of()) { 
    moly_list.append(poly_to_moly(temp->item_of())); 
  }
  int result = mixvol(moly_list,insert); 
  return result;
}

/* This function does the actual work for computing the mixed volume
   of the given list; insert is the optional flag that determines whether
   the mixed cell structure will be computed and stored or not. 

   Algorithm proceeds by computing the convex hull points for each 
   polynomial, lifting them, and then proceeding to determine which
   edges combine to form mixed cells by using the depth-first-search
   routine dfs() */
int
mixvol(List<Molynomial> moly_list,bool insert = false) 
{ 
  int num_polys, max_sites, tot_sites; 
  int *num_sites; 

  // First convert the molys into arrays of monomials, one for each moly. 
  Array<MonomArray> monoms = 
    monom_arrays(moly_list,&num_polys,&num_sites,&tot_sites,&max_sites); 

  int dim = size_of(monoms[0][0]); 
  if (dim != num_polys) { 
    cout << "Error mixvol(): Number of polynomials not equal to dimension\n";
    return 0; 
  }
  if (dim == 1) { 
    int result = max_degree(monoms[0]); 
    return result; 
  }
  double** simp_mat = make_simp_mat(dim,tot_sites); 

  Array<MonomArray> chull_pts = compute_chull_pts(monoms,simp_mat); 

  // Every point on the convex hull is lifted by an extra dimension with
  // the coordinate of the point in that dimension assigned to 1
  chull_pts = lift_chull_pts(chull_pts); 

  // Pre-allocate the edges, max number of edges per polytope is n^2 where
  // n is the number of points. The edges are stored in a three-dimensional
  // array - first dimension is the depth of computation (starting depth is
  // 0 - the depth part is required by the dfs() function), second dimension 
  // is indexed by the input polynomials, third dimension is the individual
  // hull edges. Thus edges[i][j][k] gives the k^th edge corr to the j^th
  // polytope/polynomial at the i^th level. 

  MixVolEdges Edge_info = 
    MixVolEdges(num_polys,num_sites,simp_mat,chull_pts,insert); 

  // Compute the hull edges, these are stored in edges[i] for the ith poly
  int cur_num_edges; 
  for (int i=0;i<num_polys;i++) { 
    hull_edges(chull_pts[i],Edge_info.AllEdges[0][i],&cur_num_edges,simp_mat); 
    Edge_info.AllEdges[0][i].set_size(cur_num_edges); 
    Edge_info.AllCards[0][i] = cur_num_edges; 
    // Set the size to the correct no. instead of the upper bound n^2. 
    // Bit of redundancy here - info is stored in two places in Edge_info, 
    // so as to replicate previously written code by Canny-Emiris. 
  }
  lift_all(chull_pts); 
  int mv = 0; 
  Edge_info.dfs(0,&mv,num_polys,insert); 
  return mv; 
}

/* For each moly get its monomials as an array. The returned struct is
   an array of monomial arrays, one for each moly */
Array<MonomArray> 
monom_arrays(List<Molynomial> moly_list, int* num_molys, 
int** num_sites, int *tot_sites, int* max_monoms) { 
  // First count number of molys 
  ListNode<Molynomial>* temp; 
  ListNode<Monomial>* temp2; 
  int moly_count = 0; 
  for (temp=head_of(moly_list);temp!=0;temp=temp->next_of()) { 
    moly_count++; 
  }
  Array<MonomArray> monom_mat = Array<MonomArray>(moly_count); 
  *num_sites = new int[moly_count]; 
  int cur_count = 0; int monom_count = 0; 
  int i=0; *max_monoms = 0; 
  for (temp=head_of(moly_list);temp!=0;temp=temp->next_of()) { 
    cur_count = num_monoms(temp->item_of()); 
    (*num_sites)[i] = cur_count; 
    monom_mat[i] = Array<IntPoint>(cur_count); 
    int j=0; 
    List<Monomial> cur_monoms = monoms_of(temp->item_of());
    for (temp2=head_of(cur_monoms);temp2!=0;temp2=temp2->next_of()) { 
      monom_mat[i][j] = expons_of(temp2->item_of()); 
      j++; 
    }
    i++; 
    if (cur_count > *max_monoms) 
      *max_monoms = cur_count; 
    monom_count += cur_count; 
  }
  *num_molys = moly_count; 
  *tot_sites = monom_count; 
  return monom_mat; 
}

/* Returns the matrix for use with the simplex computations. The matrix
   is pre-allocated here to its largest possible size and then used
   repeatedly for simplex computations to prevent having to re-allocate
   it each time. */
double**
make_simp_mat(int dim, int num_sites) 
{
  int num_rows = dim + num_sites + 5; 
  double** result = new (double*)[num_rows]; 
  for (int i=0;i<num_rows;i++) 
    result[i] = new double[num_sites+5]; 
  return result; 
}

/* For each array of monomials computes the corresponding convex hull
   points by invoking the function chull_points() */ 
Array<MonomArray> 
compute_chull_pts(Array<MonomArray> monoms_array,double** simp_mat)
{ 
  int num_polys = size_of(monoms_array); 
  Array<MonomArray> result = Array<MonomArray>(num_polys); 
  for (int i=0;i<num_polys;i++) { 
    result[i] = chull_points_mem(monoms_array[i],simp_mat); 
  }
  return result; 
}

/* Lifts the given arrays of points by one dimension to dim+1 dimensional
   space. The (dim+1)'st coordinate is set to 1. This coordinate is
   later set to a random value by lift_all() */ 

Array<MonomArray>
lift_chull_pts(Array<MonomArray> chull_pts)
{ 
  int num_polys = size_of(chull_pts); 
  Array<MonomArray> result = Array<MonomArray>(num_polys); 
  for (int i=0;i<num_polys;i++) { 
    result[i] = Array<IntPoint>(size_of(chull_pts[i])); 
    for (int j=0;j<size_of(chull_pts[i]);j++) { 
      int old_dim = size_of(chull_pts[i][j]);
      result[i][j] = Array<int>(old_dim+1); 
      for (int k=0;k<old_dim;k++) 
	result[i][j][k] = chull_pts[i][j][k]; 
      result[i][j][old_dim] = 1; 
    }
  }
  return result; 
}

int 
max_degree(MonomArray marray)
{ 
  if (size_of(marray[0]) != 1) { 
    cout << "Error max_degree(): called with array of non-1-dim monoms\n"; 
    return 0; 
  }
  int num_monoms = size_of(marray); 
  int result = 0; 
  for (int i=0;i<num_monoms;i++) {
    if (marray[i][0] > result) 
      result = marray[i][0]; 
  }
  return result; 
}
